/*	License
 * 		inside EU: EUPL
 * 		outside EU: GPL or EUPL
 *  The program comes AS IS and without any warranty unless required by law
 *  (C) 2012 thecodemonkey
 */

package model.transfer;

import java.io.IOException;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import net.util.Proxy;

import core.util.Observable;

import scheduler.Scheduler2;
import scheduler.SchedulerClient;
import scheduler.SchedulerRunnable2;

public abstract class ScheduledTransferManager extends Observable<TransferObserver> implements TransferManager<TransferObserver>,SchedulerClient {
	
	protected volatile boolean stopAll;
	protected Scheduler2 scheduler;
	protected long currentTotalBytes,startTimestamp;//,prevTotalBytes;
	
	protected long bytesOfFinishedTransfers;//,currentTotalSpeed;
	
	protected Timer timer;
	protected int numberOfFiles;
	protected int filesCounter;
	protected Object fileCounterLock=new Object();
	
	protected int maxThreads,maxTries = -1,timeout;
	
	protected synchronized void updateBytes() {
		
		if (!stopAll) {
			currentTotalBytes=bytesOfFinishedTransfers;
			Iterator<SchedulerRunnable2> it=scheduler.getActive().iterator();
			long currentTotalSpeed=0L;
			while (it.hasNext()) {
				TransferThread thread=(TransferThread)it.next();
				TransferEngine d=thread.engine;
				d.updateCurrentBytes();				
				TransferInfo transferInfo=thread.transferInfo;
				
				if (transferInfo.getStatus()!=Status.finished)
					transferInfo.getStatus().update(obs, transferInfo);
				currentTotalBytes+=transferInfo.getCurrentBytes();
				currentTotalSpeed+=transferInfo.getSpeed();
			}
			
			
			for (TransferObserver o:obs) 
				o.updateTotalSpeed(this,currentTotalSpeed);
			
		} 
	}
	

	protected ScheduledTransferManager(int maxThreads,Proxy proxy) {
		scheduler = new Scheduler2(this,maxThreads);
		setProxy(proxy);
		this.maxThreads = maxThreads;
	}
	
	public void setMaxThreads(int maxThreads) {
		this.maxThreads = maxThreads;
	}
	
	public void setMaxTries(int maxTries) {
		this.maxTries = maxTries;
	}
	
	protected void add(TransferEngine d) throws IOException {
		add(d, createThread(d));
	}
	
	protected void add(TransferEngine d,SchedulerRunnable2 thread) throws IOException {
		
		numberOfFiles++;
		TransferInfo di=d.getTransferInfo();
		
		if (scheduler.isRunning()){
			int id=scheduler.addTask(thread);
			di.setId(id);
			setTransferStatus(di,d, Status.queued);
		}else{
			int id=scheduler.addTaskStopped(thread);
			di.setId(id);
			setTransferStatus(di,d, Status.stopped);
		}
		for (TransferObserver o:obs) o.updateAddTransfer(di);
	}
	
	
	@Override
	public int getFileCounter() {
		return filesCounter;
	}

	@Override
	public int getNumberOfFiles() {
		return numberOfFiles;
	}

	@Override
	public long getTransferredBytes() {
		return currentTotalBytes;
	}
	
	
	public long getStartTimestamp() {
		return startTimestamp;
	}

	
	@Override
	public synchronized void stopAll() {
		if (scheduler.isRunning()) {
			cancelAll();
		}
	}
	
	protected synchronized void cancelAll() {
		stopAll=true;
		scheduler.stopAll();
		stopTimer();
		System.out.println("Request stop");
		Iterator<SchedulerRunnable2> it=scheduler.getStopped().iterator();
		
		while (it.hasNext()) {
			TransferThread thread = (TransferThread)it.next();
			TransferInfo ti = thread.transferInfo;
			ti.setLockCancel(true);
			setTransferStatus(ti,thread.engine, Status.stopped);
			ti.reset();
		}
		
		System.out.println("scheduler running "+scheduler.isRunning());
		for (TransferObserver o:obs)
			o.updateStopAllTransfers();
	}
	
	@Override
	public synchronized void startStopAll()throws IOException {
		if (scheduler.isRunning()) {
			cancelAll(); 
		}else {
			scheduler.enqueueAllStopped();
			startScheduler();
		}
	}

	public synchronized void startAll() throws IOException{
		if (!scheduler.isRunning()) {
			scheduler.enqueueAllStopped();
			startScheduler();
		}
	}
	
	
	protected void startScheduler() throws IOException{
		if (numberOfFiles>0) {
			stopAll=false;
			scheduler.setMaxThreads(maxThreads);
			TransferEngine.setTimeout(timeout);
			
			Iterator<SchedulerRunnable2> it=scheduler.getUnfinished().iterator();
			
			while (it.hasNext()) {
				TransferThread thread = (TransferThread) it.next();
				TransferInfo info = thread.transferInfo;
				if (info.getStatus()!=Status.finished) {
					info.setLockCancel(false);
					setTransferStatus(info, thread.engine, Status.queued); 
				}
			}
			startTimer();
			scheduler.start();
		}
	}
	
	protected void stopTimer() {
		if (timer!=null){
			System.out.println("timer "+timer);
			timer.cancel();
			timer =null;
		}
	}
	
	protected void startTimer() {
		System.out.println("Start timer");
		if (timer == null) {
			timer=new Timer();
			timer.scheduleAtFixedRate(new TimerTask() {
				
				@Override
				public void run() {
					updateBytes();
				}
			}, 1000,1000);
		}
		
	}
	
	@Override
	public void stopEnqueue(int id) throws IOException{
		if (id>-1) {
			if (scheduler.isRunning(id)) {
				scheduler.stop(id);
				if (!scheduler.isRunning()) stopTimer();
			}else {
				scheduler.enqueueStopped(id);
				
				if (!scheduler.isRunning()) startScheduler();
			}
		}
	}
	
	protected abstract SchedulerRunnable2 createThread(TransferEngine d) ;
	
	protected abstract class TransferThread extends SchedulerRunnable2 {

		protected TransferEngine engine;
		protected TransferInfo transferInfo;
		
		public TransferThread(TransferEngine d) {
			this.engine=d;
			transferInfo=d.getTransferInfo();
		}
		
		protected abstract void runAction();
		
		@Override
		public void enqueue() {
			transferInfo.setLockCancel(false);
			setTransferStatus(transferInfo,engine,Status.queued);
		}
		
		@Override
		public void run() {
			stop=false;
			do {
				transferInfo.reset();
				runAction();
			} while (transferInfo.getStatus()!=Status.finished &&transferInfo.getError()==null && !stop);
		}

		@Override
		public void stop() {
			if (engine!=null) {
				stop=true;
				setTransferStatus(transferInfo,engine, Status.stopped);
				engine.cancel();
				transferInfo.reset();
			}
		}
		
		public TransferInfo getTransferInfo() {
			return transferInfo;
		}
		
		
		public TransferEngine getEngine() {
			return engine;
		}

		public void setEngine(TransferEngine engine) {
			this.engine = engine;
		}
		
		public void setTransferInfo(TransferInfo transferInfo) {
			this.transferInfo = transferInfo;
		}
		

		@Override
		public boolean errorOccoured() {
			return transferInfo.getError()!=null;
		}
	}
	

	@Override
	public void removeAll() {
		stopAll();
		for (TransferObserver o:obs) {
			o.updateRemoveAll();
		}
		scheduler.removeAll();
		numberOfFiles=0;
		filesCounter=0;
		bytesOfFinishedTransfers=0;
		currentTotalBytes=0;
//		prevTotalBytes=0;
	}
	
	@Override
	public void remove(int id) {
		TransferThread thread= (TransferThread) scheduler.remove(id);
		setTransferStatus(thread.transferInfo,thread.engine, Status.remove);
		numberOfFiles--;
	}

	protected void setTransferStatus(TransferInfo transferInfo,TransferEngine engine, Status status) {
		transferInfo.setStatus(status);
		status.update(obs, transferInfo);
	}


	@Override
	public void setTimeout(int timeout) {
		this.timeout=timeout;		
	}


	@Override
	public void setProxy(Proxy proxy) {
		TransferEngine.setProxy(proxy);
	}
	
	@Override
	public List<TransferInfo> getUnfinished() {
		LinkedList<TransferInfo> inf=new LinkedList<>();
		
		Collection<SchedulerRunnable2> threads=scheduler.getUnfinished();
		for (SchedulerRunnable2 t:threads) {
			inf.add( ((TransferThread)t).transferInfo );
		}
		
		return inf;
	}
}
